/**
 * @file FSRegistry.cpp
 * @author apio (cloudapio.eu)
 * @brief Kernel registry of available file system implementations.
 *
 * @copyright Copyright (c) 2023, the Luna authors.
 *
 */

#include "fs/FSRegistry.h"
#include "fs/devpts/FileSystem.h"
#include "fs/ext2/FileSystem.h"
#include "fs/tmpfs/FileSystem.h"
#include "lib/Mutex.h"

struct VirtualFileSystemDescriptor
{
    StringView name;
    VirtualFileSystemFactoryFunction factory_function;
};

struct PhysicalFileSystemDescriptor
{
    StringView name;
    PhysicalFileSystemFactoryFunction factory_function;
};

Vector<VirtualFileSystemDescriptor> g_virtual_filesystems;
Vector<PhysicalFileSystemDescriptor> g_physical_filesystems;

static Mutex g_lock;

namespace FSRegistry
{
    Result<void> init()
    {
        TRY(register_virtual_filesystem(TmpFS::FileSystem::create, "tmpfs"));
        TRY(register_virtual_filesystem(DevPTS::FileSystem::create, "devpts"));
        TRY(register_virtual_filesystem(DeviceRegistry::create_devfs_instance, "devfs"));
        TRY(register_physical_filesystem(Ext2::FileSystem::create, "ext2"));

        return {};
    }

    Result<void> register_virtual_filesystem(VirtualFileSystemFactoryFunction factory_function, StringView name)
    {
        ScopedMutexLock lock(g_lock);

        for (const auto& descriptor : g_virtual_filesystems)
        {
            if (descriptor.name == name) return err(EEXIST);
        }

        auto descriptor = VirtualFileSystemDescriptor { .name = name, .factory_function = factory_function };

        return g_virtual_filesystems.try_append(move(descriptor));
    }

    Result<void> register_physical_filesystem(PhysicalFileSystemFactoryFunction factory_function, StringView name)
    {
        ScopedMutexLock lock(g_lock);

        for (const auto& descriptor : g_physical_filesystems)
        {
            if (descriptor.name == name) return err(EEXIST);
        }

        auto descriptor = PhysicalFileSystemDescriptor { .name = name, .factory_function = factory_function };

        return g_physical_filesystems.try_append(move(descriptor));
    }

    Result<VirtualFileSystemFactoryFunction> lookup_virtual_filesystem(StringView name)
    {
        ScopedMutexLock lock(g_lock);

        for (const auto& descriptor : g_virtual_filesystems)
        {
            if (descriptor.name == name) return descriptor.factory_function;
        }

        return err(ENODEV);
    }

    Result<PhysicalFileSystemFactoryFunction> lookup_physical_filesystem(StringView name)
    {
        ScopedMutexLock lock(g_lock);

        for (const auto& descriptor : g_physical_filesystems)
        {
            if (descriptor.name == name) return descriptor.factory_function;
        }

        return err(ENODEV);
    }
}
